/*
** showtime -- movie viewer
** Display a series of frames rapidly, to create the illusion of animation.
**
** Main movie loop:
**  This module is a collection point of all other modules:  It is the
**  main control module.  From here, playback is controlled, and user
**  input is disseminated to other modules.
**
**  Since this is a collection point, global variables should not be
**  defined here, since routines in other modules should be called by
**  routines in this module, not vice versa.
**
**  (One exception is gm_in_do_xform, which is global so that the x-video
**  display routine can hold off displaying the image to the screen until
**  after all image transformations have been performed.  This reduces
**  flicker and scrolling.)
**
** Copyright 1992, 1993, 1994 by
** Michael J. Gourlay
**
** NO WARRANTEES, EXPRESS OR IMPLIED.
*/

extern int pgm_cols_title;
extern int pgm_rows_title;
extern int pgm_maxval_title;
extern unsigned char pgm_image_title[];

#include <fcntl.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "showtime.h"
#include "image.h"
#include "cmap.h"
#include "pgmcload.h"
#include "video.h"
#include "file.h"
#include "font.h"
#include "pause.h"

#include <stdarg.h>

#define USAGE \
"   ShowTime: display a series of images rapidly\n\
   Copyright (C) 1993, 1994 by Michael J. Gourlay and Georgia Tech\n\
 usage: showtime [filename] [options]\n\
   options:\n\
     [-f] <filename>     : script filename\n\
     -c                  : force loading CGA palette in first 16 slots\n\
     -n <saturation>     : rescale saturation of palette\n\
     -s <delay>          : step mode;  pause between each frame\n\
     -w <width> [height] : suggest a minimum window size\n"


/* ---------------------------------------------------------------------- */
/* GLOBAL VARIABLES */

/* flag to tell whether we're in a do_xform */
int gm_in_do_xform=0;

/* ---------------------------------------------------------------------- */
/* variables scoped to this module */

/* FIOD is an array of files used for viewing movies */
/* There are MAX_FIODS+1 files and the 0th file is the default */
/* image.c uses this */
#define MAX_FIODS    10
static FILE *FIOD[MAX_FIODS+1];

/* screendata initializer record */
static const screendata_t gm_sd_init = {
  0,    /* header_size */
  0,    /* num_rows */
  0,    /* num_cols */
  0,    /* xoffset */
  0,    /* yoffset */
  NULL, /* data */
};

/* gm_step is the inter-image delay in seconds */
static float gm_step=0;

/* playback mode */
static play_mode_t gm_play_mode;

/* log file pointer */
static FILE *logfp=NULL;

/* gm_xform_script is the list of xform commands */
#define MAX_XFORM_CMDS 100
static char *gm_xform_script[MAX_XFORM_CMDS];

/* Image data for drawing text on the screen */
static screendata_t font_sd;

/* List of file pointer locations in streams */
/* This is a list of where within a file an image starts (startframes) */
static long gm_img_loc[MAX_FIODS+1][MAX_IMAGES];

/* gm_img_loc_index is the index for the location of the current image */
static int gm_img_loc_index[MAX_FIODS+1];

/* ====================================================================== */

/* bye_bye: Clean up and exit with value rv */
void
bye_bye(int rv)
{
  reset_display();
  if(logfp!=NULL) fclose(logfp);
  exit(rv);
}

/* ====================================================================== */

/* logmesg: Error message logging (to a file) from vprintf args */
void
logmesg(char *fmt, ...)
{
  char *log_filename="showtime.log";
  va_list ap;
  va_start(ap, fmt);
  if(logfp==NULL) {
    /* Open a log file to output messages instead of messing up screen */
    if((logfp=fopen(log_filename, "wc"))==NULL) {
      /* This is the only place where stderr should be used */
      /* use logfp for error messages elsewhere */
      fprintf(stderr, "showtime: could not open log file %s for writing\n",
              log_filename);
      fprintf(stderr, "the message attempted to be displayed was:\n");
      vfprintf(stderr, fmt, ap);
      bye_bye(1);
    }
  }
  vfprintf(logfp, fmt, ap);
  va_end(ap);
  fflush(logfp);
}

/* ====================================================================== */

#define UBYTE    unsigned char

/* init_movie: allocate memory for the row pointers in sdP
** Set each row pointer to NULL so that we know that they're empty.
** This must be guarenteed for other routines to trigger off NULL pointers.
** Does not allocate memory for image space, only pointers to rows.
**
** Return nonzero if malloc fails
** Return zero if everything works okay
*/
int
init_movie(int num_rows, int num_cols, screendata_t *sdP)
{
  register int row;

  sdP->num_rows = num_rows;
  sdP->num_cols = num_cols;

  sdP->data = (UBYTE **)malloc(sizeof(UBYTE **) * (size_t)sdP->num_rows);
  if(sdP->data==NULL) {
    logmesg("\ninit_movie: could not allocate %i pointers\n",
            sdP->num_rows);
    reset_display();
    return(1);
  }

  /* Initialize all row pointers to NULL so we know they're empty */
  for(row=0; row < (sdP->num_rows); row++)
    sdP->data[row]=NULL;

  return(0);
}


/* ---------------------------------------------------------------------- */

/* free_movie : free allocated memory for screen data sdP
*/
void
free_movie(screendata_t *sdP)
{
  register int row;

  if(sdP->data==NULL) return;

  for(row=0; row<sdP->num_rows; row++)
    if(sdP->data[row]!=NULL) {
      free(sdP->data[row]);
    }

  if(sdP->data!=NULL) {
    free(sdP->data);
  }
}

/* ====================================================================== */

/* draw_text: draw a string str as bitmapped font characters on the screen */
static void
draw_text(char *str)
{
  int sindx;      /* string index */
  int fcxp, fcyp; /* font char position */
  int oldx, oldy; /* temp space to remember global image offset positions */

  /* initialize the screen positions of the string */
  fcxp=0; fcyp=0;

  /* Draw each character, each to the right of the previous */
  for(sindx=0; (size_t)sindx<strlen(str); sindx++) {
    /* load font images from memory */
    load_pgm_from_c(pgm_cols_font_char[str[sindx]],
      pgm_rows_font_char[str[sindx]],
      &font_sd, pgm_image_font_char[str[sindx]]);

    /* remember the old offset */
    oldx=gm_xoff; oldy=gm_yoff;

    /* set the offset to the location of the char string */
    gm_xoff+=fcxp; gm_yoff+=fcyp;

    /* display the char string */
    show_image(&font_sd);

    /* update the position of the next letter in the string */
    fcxp += pgm_cols_font_char[str[sindx]]*gm_xscale;

    /* restore the old offset */
    gm_xoff=oldx; gm_yoff=oldy;
  }
}

/* ---------------------------------------------------------------------- */

/* handle_events: handle mouse and keyboard events
** Playback Actions:
**   pause unpause reverse forward step stop
*/
static play_mode_t
handle_events(play_mode_t play_mode)
{

#if defined(X11)

  XEvent event;

  if(XCheckMaskEvent(x_display, x_event_mask, &event)) {
    if(event.type == KeyPress) {
      char key;
      if(XLookupString(&event.xkey, &key, 1, NULL, NULL)) {

#elif defined (MSC)

      int key;

      if(kbhit()) {
        key=getch();
#endif

        switch(key) {

          /* Pause and step */
          case ' ':
            if(play_mode.play == MODE_PLAY) {
              play_mode.step=MODE_STEP;
            }
            play_mode.play=MODE_PAUSE;
            play_mode.commands=MODE_NO_COMMANDS;
          break;

          /* resume play */
          case 'p': case 'P': case 10 /*nl*/: case 13 /*cr*/:
            play_mode.play=MODE_PLAY;
            play_mode.commands=MODE_COMMANDS;
          break;

          /* resume commands */
          case ':': case 'c': case 'C':
            /*play_mode.play=MODE_PAUSE;/**/
            play_mode.commands=MODE_COMMANDS;
          break;

          /* stop */
          case 'q': case 'Q': case 27 /*esc*/:
            play_mode.stop=MODE_STOP;
          break;

          /* Reverse direction */
          case ',': case 'r':
            play_mode.direction=MODE_REVERSE;
            if(play_mode.play==MODE_PAUSE) play_mode.step=MODE_STEP;
          break;

          /* Reverse direction, pause playback, accept commands */
          case '<': case 'R':
            play_mode.direction=MODE_REVERSE;
            play_mode.play=MODE_PAUSE;
            play_mode.step=MODE_STAY;
            play_mode.commands=MODE_COMMANDS;
          break;

          /* Forward direction */
          case '.': case 'f':
            play_mode.direction=MODE_FORWARD;
            if(play_mode.play==MODE_PAUSE) play_mode.step=MODE_STEP;
          break;

          /* Forward direction, pause playback, accept commands */
          case '>': case 'F':
            play_mode.direction=MODE_FORWARD;
            play_mode.play=MODE_PAUSE;
            play_mode.step=MODE_STAY;
            play_mode.commands=MODE_COMMANDS;
          break;

#if defined(MSC)
          case 0: case 0xe0:
            /* Keypress was a function key, so get next char */
            getch();
          break;
#endif /* MSC */

          default:
          break;

        } /* switch(key) */
      } /* } */ /* if(XLookupString) OR if(kbhit()) */

#if defined(X11)
    } else if(event.type == ButtonPress) {
      switch(event.xbutton.button) {
        case 1: /* left button */
          play_mode.direction=MODE_REVERSE;
          if(play_mode.play==MODE_PAUSE) play_mode.step=MODE_STEP;
        break;

        case 2: /* middle button */
          if(play_mode.play == MODE_PLAY) {
            play_mode.play=MODE_PAUSE;
            play_mode.step=MODE_STEP;
            play_mode.commands=MODE_NO_COMMANDS;
          } else {
            play_mode.play=MODE_PLAY;
          }
        break;

        case 3: /* right button */
          play_mode.direction=MODE_FORWARD;
          if(play_mode.play==MODE_PAUSE) play_mode.step=MODE_STEP;
        break;

        default:
        break;
      } /* switch(event.xbutton.button) */
    } /* if(event.type) */
  } /* if(XCheckEventMask) */
#endif

  return(play_mode);
}

/* ---------------------------------------------------------------------- */

static int
recall_image(int fiod_index, int image_index)
{
  /* The last known image location is followed by a NULL */
  if(gm_img_loc[fiod_index][image_index] == -1) {
    return(EOF);
  }

  /* Go to where the beginning of the image is */
  return(fseek(FIOD[fiod_index], gm_img_loc[fiod_index][image_index], 0));
}

/* ---------------------------------------------------------------------- */

/* remember_image: advance frame counter and move to startframe */
static int
remember_image(int fiod_index, play_mode_t play_mode)
{
  if(play_mode.direction == MODE_REVERSE) {
    /* Recall where the header starts (startframe) */
    if(recall_image(fiod_index, gm_img_loc_index[fiod_index]-1)) {
      return(EOF);
    } else {
      /* Reverse the index to the previous image location */
      gm_img_loc_index[fiod_index]--;
    }
  } else {
    /* Advance the index to the next image location */
    gm_img_loc_index[fiod_index]++;

    /* Keep track of the end of known image locations */
    /* The last known image location is followed by a -1 */
    if(gm_img_loc[fiod_index][gm_img_loc_index[fiod_index]] == -1) {
      gm_img_loc[fiod_index][gm_img_loc_index[fiod_index]+1] = -1;
    }

    /* Remember where the beginning of this image is */
    gm_img_loc[fiod_index][gm_img_loc_index[fiod_index]]=
      ftell(FIOD[fiod_index]);

    /* If ftell returned -1, there is a file error */
    if(gm_img_loc[fiod_index][gm_img_loc_index[fiod_index]] == -1) {
      gm_img_loc_index[fiod_index]--;
      return(-1);
    }

    /* If previous image started at same place, do not count this image */
    if(gm_img_loc[fiod_index][gm_img_loc_index[fiod_index]] ==
       gm_img_loc[fiod_index][gm_img_loc_index[fiod_index]-1]) {
      gm_img_loc[fiod_index][gm_img_loc_index[fiod_index]] = -1;
      gm_img_loc_index[fiod_index]--;
      return(EOF);
    }
  }
  return(0);
}

/* ---------------------------------------------------------------------- */

/* preload_movie: find all startframes in a file of images and move to last
** Pass in the fiod_index.
** The element in the global list of open files, FIOD[fiod_index],
** must be set to the opened file.
** The file descriptor will be pointing to the last startframe, and the fseek
** location array gm_img_loc (startframe list) will be filled, and the
** index into that startframe array (gm_img_loc_index[fiod_index] will
** be set to the last image in the list.
*/
static int
preload_movie(int fiod_index)
{
  screendata_t sd = gm_sd_init;
  play_mode_t  pm = gm_play_mode;
  int   ri_return;

  pm.direction = MODE_FORWARD;

  /* Read image file */
  do {
    /* Remember where the header starts (startframe) */
    if(remember_image(fiod_index, pm)) {
      return(EOF);
    }
    /* skip through the movie image */
    ri_return=load_image(&sd, FIOD[fiod_index]);
  } while(!ri_return);

  free_movie(&sd);

  /* Move back one frame to the startframe of the last frame */
  gm_img_loc_index[fiod_index]--;
  if(recall_image(fiod_index, gm_img_loc_index[fiod_index])) {
    return(EOF);
  }

  return(0);
}

/* ---------------------------------------------------------------------- */

/* free_block: free block of commands in cmd_list
*/
static void
free_block(char *cmd_list[])
{
  int  cmd_index;

  for(cmd_index=0; cmd_list[cmd_index]!=NULL; cmd_index++) {
    free(cmd_list[cmd_index]);
    cmd_list[cmd_index] = NULL;
  }
}

/* ---------------------------------------------------------------------- */

/* copy_block: copy block of commands from cmd_list or stdin
** if cmd_list is NULL, read from stdin
** return line number of end_cmd or negative of line number of NULL command
*/
static int
copy_block(int start_line, char *end_cmd, char *cmd_list[], char *dlist[])
{
  int  cmd_index=start_line, dlist_index=0;
  char cmd_line[256];
  char cmd[256];
  int  scan_items;
  int  done=FALSE;

  while(!done) {
    /* Read a command line */
    if(cmd_list!=NULL) {
      /* Read command from script in memory */
      if(cmd_list[cmd_index]!=NULL) {
        strcpy(cmd_line, cmd_list[cmd_index]);
      } else {
        /* reached end of list before end command */
        done=-1;
      }

    } else {
      /* Read command from file */
      if(fgets(cmd_line, 255, stdin)==NULL) {
        done=-1;
      }
    }
    cmd_index++;

    /* Extract the command string from the command line */
    scan_items=sscanf(cmd_line, "%s", cmd);
    if(scan_items != 1) cmd[0]='\0';

    /* Store the command line */
    if((dlist[dlist_index]=strdup(cmd_line))==NULL) {
      logmesg("could not allocate block line %i\n", dlist_index);
      bye_bye(1);
    }
    if(STREQ(cmd, end_cmd)) {
      done=1;
    }
    ++dlist_index;
  }
  dlist[dlist_index]=NULL;

  if(done<0) return(-(cmd_index-1));
  else return(cmd_index-1);
}

/* ---------------------------------------------------------------------- */

/* pre_parse: for reverse-direction command parsing */
int
pre_parse(char *cmd, char* cmd_line, char **cmd_script, int *cmd_index)
{
  /* If we found an :endxform, load the _previous_ xform, or empty if none */
  if(STREQ(":endxform", cmd)) {
    int ci, scan_items;
    char cs[256];
    int first_xform=*cmd_index;
    int xforms_found = 0;

    /* Search backwards for ":xform" */
    for(ci=*cmd_index; ci>=0; ci--) {
      scan_items=sscanf(cmd_script[ci], "%s", cs);
      if(scan_items<1) cs[0]='\0';

      if(STREQ(":xform", cs)) {
        xforms_found++;
        if(xforms_found>1) {
          if(gm_xform_script[0]!=NULL) {
            free_block(gm_xform_script);
          }
          /* Skip the begin-block command */
          ci++;
          /* Copy in the new xform */
          copy_block(ci, ":endxform", cmd_script, gm_xform_script);
          /* flag end of loop */
          ci = -1;
        } else {
          /* Remember where the first xform was so we can point to it later */
          first_xform = ci;
        }
      }
    }
    if(xforms_found < 2) {
      /* No previous xform found, so empty the xform script */
      if(gm_xform_script[0]!=NULL) {
        free_block(gm_xform_script);
      }
    }
    *cmd_index=first_xform-1;
    /* dummy up command */
    cmd[0] = cmd_line[0] = '\0';
  }
  return(*cmd_index);
}

/* ---------------------------------------------------------------------- */

/* parse_simple: parse command cmd with args cmd_line
** return 1 if cmd was not among these simple commands
** return 0 if cmd ran okay
** return EOF if EOF was reached reading from the movie images
*/
static int
parse_simple(char *cmd, char *cmd_line, int cmd_len, screendata_t *sdP)
{
  int scan_items;
  int fiod_index=0;

  /* Parse these in order of commands most likely to be executed
  ** so the most likely are executed faster
  */
  if(STREQ(":next", cmd) || STREQ(":load", cmd)) {
    /* Load (and maybe display) the next image in an already open file */
    scan_items=sscanf(cmd_line+cmd_len, "%i", &fiod_index);
    if(scan_items<1) fiod_index=0;

    if(fiod_index<0) fiod_index=0;
    if(fiod_index>=MAX_FIODS) {
      fiod_index=MAX_FIODS;
      logmesg("%s -- can use up to %i files, using last one\n", cmd, MAX_FIODS);
    }

    if(! FIOD[fiod_index]) {
      logmesg("%s -- movie file %i is not open.\n", cmd, fiod_index);
      return(EOF);
    } else {

      /* Remember where the header starts (startframe) */
      if(remember_image(fiod_index, gm_play_mode)) {
        return(EOF);
      }

      /* Read and display image */
      scan_items=load_image(sdP, FIOD[fiod_index]);
      if(scan_items) {
        /* Reached end of file */
        if(gm_play_mode.direction == MODE_FORWARD) {
          gm_img_loc[fiod_index][gm_img_loc_index[fiod_index]] = -1;
          gm_img_loc_index[fiod_index]--;
        }
        /* Tell the caller that we reached end of file */
        return(EOF);
      } else {
        if(STREQ(":next", cmd)) {
        /* Start the pause timer */
        if(gm_step!=0.0) pause_init();
          show_image(sdP);
        }
        if(gm_step!=0.0) pause_finish(gm_step);
      }
    }

  } else if(STREQ(":repeat", cmd) || STREQ(":show", cmd)) {
    /* Display image */
    if(gm_step!=0.0) pause_init();
    show_image(sdP);
    if(gm_step!=0.0) pause_finish(gm_step);

  } else if(STREQ(":frame", cmd)) {
    /* Display the frame number */
    char frame_s[10];

    scan_items=sscanf(cmd_line+cmd_len, "%i", &fiod_index);
    if(scan_items<1) fiod_index=0;

    /* Create the string */
    sprintf(frame_s, " %04i ", gm_img_loc_index[fiod_index]);

    draw_text(frame_s);

  } else if(STREQ(":vflip", cmd)) {
    /* Vertically flip an already loaded image */
    image_vflip(sdP);

  } else if(STREQ(":hflip", cmd)) {
    /* Horizontally flip an already loaded image */
    image_hflip(sdP);

  } else if(STREQ(":offset", cmd)) {
    /* Set the position of where the image UL corner is on the screen */
    scan_items=sscanf(cmd_line+cmd_len, "%i %i", &gm_xoff, &gm_yoff);
    if(scan_items<1) gm_xoff=gm_yoff=0; else if(scan_items<2) gm_yoff=0;

  } else if(STREQ(":scale", cmd)) {
    /* Set the scale factors of the image */
    scan_items=sscanf(cmd_line+cmd_len, "%f %f", &gm_xscale, &gm_yscale);
    if(scan_items<1) gm_xscale=gm_yscale=1; else if(scan_items<2) gm_yscale=1;
    if(gm_xscale==0) gm_xscale=1;
    if(gm_yscale==0) gm_yscale=1;

  } else if(STREQ(":rem", cmd) || cmd[0]=='#') {
    /* Comment line */

  } else if(STREQ(":rapid", cmd)) {
    /* Rapid-play a movie image file and close when done */
    /* "rapid" means ignore keyboard and mouse events */
    /* Also, the :frame command will not work since no FIOD is used */
    char  movie_filename[256];
    int   ri_return;
    FILE *fio;

    scan_items=sscanf(cmd_line+cmd_len, "%s", movie_filename);
    if(scan_items<1) strcpy(movie_filename, "movie");

    /* open image file */
    if((fio=open_file(movie_filename))==NULL) {
      logmesg("\n%s: Unable to open file \"%s\".\n", cmd, movie_filename);
      return(1);
    }

    /* display image file */
    do {
      if(gm_step!=0.0) pause_init();
      ri_return=load_image(sdP, fio);
      show_image(sdP);
      pause_finish(gm_step);
    } while(!ri_return);

    if(fio) fclose(fio) ;

  } else if(STREQ(":echo", cmd)) {
    /* Draw a string of text characters on the display */
    char *str;

    str=cmd_line+cmd_len;
    if(str[0] == 32) str++;
    if(strlen(str)) draw_text(str);

  } else if(STREQ(":palette", cmd)) {
    /* Load and install a palette */
    char palette_filename[256];
    unsigned char cmap[256][3];

    scan_items=sscanf(cmd_line+cmd_len, "%s", palette_filename);
    if(scan_items<1) strcpy(palette_filename,"palette.ppm");
    if(STREQ(":rgb332", palette_filename)) {
      make_palette(RGB332, cmap);
      install_cmap(256, cmap, 255);
    } else if(STREQ(":gray", palette_filename)) {
      make_palette(GRAY, cmap);
      install_cmap(256, cmap, 255);
    } else {
      load_palette_file(palette_filename);
    }

  } else if(STREQ(":pause", cmd)) {
    /* Pause */
    float pause_time;

    scan_items=sscanf(cmd_line+cmd_len, "%f", &pause_time);
    if(scan_items<1) pause_time=-5.0;
    st_pause(pause_time);

  } else if(STREQ(":clear", cmd)) {
    /* Clear the screen, to a pixel index if provided, 0 otherwize */
    int bg_color;

    scan_items=sscanf(cmd_line+cmd_len, "%i", &bg_color);
    if(scan_items<1) bg_color=0;
    if(bg_color<0) bg_color = 0;
    else if(bg_color>255) bg_color = 255;

    clear_display(bg_color);

  } else if(STREQ(":open", cmd)) {
    /* Open a movie image file */
    char movie_filename[256];

    /* Find the file name and index */
    scan_items=sscanf(cmd_line+cmd_len, "%s %i", movie_filename, &fiod_index);
    if(scan_items<1) {
      logmesg("%s needs a filename\n", cmd);
      return(1);
    }
    if(scan_items<2) fiod_index=0;

    if(fiod_index<0) fiod_index=0;
    if(fiod_index>=MAX_FIODS) {
      fiod_index=MAX_FIODS;
      logmesg("can :open up to %i files, using last one\n", MAX_FIODS);
    }

    /* close image file if it was open */
    if(FIOD[fiod_index]) {
      fclose(FIOD[fiod_index]);
    }

    /* open image file, get file pointer, reset startframe list */
    if(( FIOD[fiod_index]=open_file(movie_filename)) == NULL) {
      logmesg("parse_simple: failed to :open file %s\n", movie_filename);
      return(EOF);
    }

    /* Empty out the list of image startframe */
    if(fiod_index >= 0) {
      int img_indx;
      gm_img_loc_index[fiod_index] = 0;
      for(img_indx=0; img_indx< MAX_IMAGES; img_indx++) {
        gm_img_loc[fiod_index][img_indx] = -1;
      }
    }

    /* If we're running backwards, we need to preload the startframes */
    if(gm_play_mode.direction == MODE_REVERSE) {
      preload_movie(fiod_index);
    }

  } else if(STREQ(":close", cmd)) {
    /* Close a movie image file */
    scan_items=sscanf(cmd_line+cmd_len, "%i", &fiod_index);
    if(scan_items<1) fiod_index=0;

    if(fiod_index<0) fiod_index=0;
    if(fiod_index>=MAX_FIODS) {
      fiod_index=MAX_FIODS;
      logmesg("%s -- can use up to %i files.\n", cmd, MAX_FIODS);
    }

    if(! FIOD[fiod_index]) {
      logmesg("%s -- movie file %i is not open.\n", cmd, fiod_index);
    } else {
      fclose(FIOD[fiod_index]);
      FIOD[fiod_index]=NULL;
    }

  } else if(STREQ(":author", cmd)) {
    /* Display author information */
    /* load title image from memory but (do not display) */
    load_pgm_from_c(pgm_cols_title, pgm_rows_title, sdP, pgm_image_title);
    logmesg("ShowTime\n");
    logmesg("by Michael J. Gourlay\n");
    logmesg("version %s\n", ST_VERSION);

  } else if(STREQ(":ntsc", cmd)) {
    /* Set the saturation scale factor to resaturate the next palette */
    scan_items=sscanf(cmd_line+cmd_len, "%f", &gm_ntsc);
    if(scan_items<1) gm_ntsc=1.0;

  } else if(STREQ(":resume", cmd)) {
    gm_play_mode.commands=MODE_NO_COMMANDS;

  } else if(STREQ(":forward", cmd)) {
    gm_play_mode.direction=MODE_FORWARD;

  } else if(STREQ(":reverse", cmd)) {
    gm_play_mode.direction=MODE_REVERSE;

  } else {
    return(1);
  }

  return(0);
}

/* ---------------------------------------------------------------------- */

/* exec_block: execute block of commands
** return line number of end_cmd or negative of line number of NULL command
*/
static int
exec_block(int start_line, char *end_cmd, char *cmd_list[], screendata_t *sdP)
{
  int  cmd_index;
  char cmd_line[256];
  char cmd[256];
  int  cmd_len;
  int  scan_items;
  int  done=FALSE;

  cmd_index=start_line;
  while(cmd_list[cmd_index]!=NULL && (!done)) {
    /* Read command from cmd_list */
    strcpy(cmd_line, cmd_list[cmd_index]);

    /* Extract the command string from the command line */
    scan_items=sscanf(cmd_line, "%s%n", cmd, &cmd_len);
    if(scan_items>0) { /* ignore blank lines */
      if(STREQ(cmd, end_cmd)) {
        done=TRUE;
      } else if(parse_simple(cmd, cmd_line, cmd_len, sdP)==1) {
        logmesg("exec_block: ignoring nonsimple command [%i] '%s'\n",
          cmd_index, cmd_line);
      }
    }
    cmd_index++;
  }

  if(!done) return(-(cmd_index-1));
  else return(cmd_index-1);
}

/* ---------------------------------------------------------------------- */

/* do_xform: execute "command macro" on image sdP */
void
do_xform(screendata_t *sdP)
{
  int  xform_index;

  gm_in_do_xform++;

  if((xform_index=exec_block(0, ":endxform", gm_xform_script, sdP))<0) {
    logmesg("do_xform: unexpected end of script %i\n", xform_index);
  }

  if(gm_in_do_xform>0) gm_in_do_xform--; else gm_in_do_xform=0;
}

/* ====================================================================== */

/* load_text_file: load text file from in_fp into text as a list of strings */
static int
load_text_file(FILE *in_fp, char ***text)
{
  int  cmd_line_count;
  char cmd_line[256];

  /* Count the number of lines in the command file */
  cmd_line_count=0;
  while(fgets(cmd_line, 255, in_fp)) {
    cmd_line_count++;
  }

  /* Reset the file pointer to the beginning of the command file */
  fseek(in_fp, 0, 0);

  /* Allocate the array of string pointers for the internal command list */
  if((*text=(char **)malloc(sizeof(char *)*(size_t)(cmd_line_count+2)))==NULL) {
    logmesg("load_text_file: could not allocate bytes\n");
    return(1);
  }

  /* Read the command file into memory */
  cmd_line_count=0;
  while(fgets(cmd_line, 255, in_fp)) {
    if( ((*text)[cmd_line_count] = strdup(cmd_line)) == NULL) {
      logmesg("load_text_file: failed to allocate memory\n");
      bye_bye(1);
    }
    cmd_line_count++;
  }
  /* Make the last two lines NULL as end-of-list markers */
  (*text)[cmd_line_count] = NULL;
  (*text)[cmd_line_count+1] = NULL;

  /* Reset the file pointer to the beginning of the command file */
  fseek(in_fp, 0, 0);

  return(cmd_line_count);
}

/* ====================================================================== */
/* ---------------------------------------------------------------------- */
/* ====================================================================== */

/* Command line argument parse macro */
#define cl_match_arg(cmd) (!strcmp(argv[apc], cmd) && ((apc+1)<argc))

int
main(int argc, char **argv)
{
  char         cmd_line[256];
  char         cmd[256];
  int          cmd_len;

  char         *cmd_filename = NULL;
  char         **cmd_script = NULL;
  char         cmd_script_active = FALSE;
  int          cmd_direction = 1;
  int          cmd_index = -1;
  FILE         *in_fp;

  int          scan_items;
  int          indx;
  int          r_width=320, r_height=200; /* requested window size */
  int          no_title=0; /* whether to skip the title screen */
  screendata_t sd = gm_sd_init;

  int          apc;

  /* Initialize variables */
  for(indx=0; indx<=MAX_FIODS; indx++) {
    FIOD[indx]=NULL;
  }

  gm_play_mode.play=MODE_PAUSE;
  gm_play_mode.step=MODE_STAY;
  gm_play_mode.direction=MODE_FORWARD;
  gm_play_mode.commands=MODE_COMMANDS;
  gm_play_mode.stop=MODE_GO;

  memset(gm_xform_script, 0, sizeof(char *) * MAX_XFORM_CMDS);

  /* Parse command line arguments */
  for(apc=1; apc<argc; apc++) {

    if(argv[apc][0] != '-') {
      /* Use this argument as the script filename */
      if(cmd_filename==NULL) {
        cmd_filename=argv[apc];
      } else {
        fprintf(stderr, USAGE);
        logmesg("bad option: %s\n", argv[apc]);
        logmesg(USAGE);
      }
    } else {
      switch(argv[apc][1]) {

        case 'f': /* command file : filename */
          if(cl_match_arg("-f")) {
            cmd_filename=argv[++apc];
          } else {
            logmesg("%s requires a filename\n", argv[apc]);
          }
          break;

        case 'c': /* CGA palette */
          gm_force_cga=1;
          break;

        case 'n': /* NTSC saturation factor */
          gm_ntsc=0.7;
          break;

        case 's': /* step mode : delay in seconds */
          if(cl_match_arg("-s")) {
            scan_items=sscanf(argv[++apc], "%f", &gm_step);
            if(scan_items<1) {
              /* no numeric arg means keypress delay with a 1 minute max */
              gm_step = -60.0;
              apc--;
            }
          } else {
            /* step w/o time argument */
            /* no arg means keypress delay with a 1 minute max */
            gm_step = -60.0;
          }
          break;

        case 'w': /* window size : width height */
          if(cl_match_arg("-w")) {
            scan_items=sscanf(argv[++apc], "%i", &r_width);
            if(scan_items<1) {
              /* no numeric args, complain */
              apc--;
              logmesg("%s requires an integer\n", argv[apc]);
            } else {
              if(apc+1 < argc) {
                scan_items=sscanf(argv[++apc], "%i", &r_height);
                if(scan_items<1) {
                  /* one numeric arg, guess other */
                  r_height = 3*r_width/4;
                  apc--;
                }
              } else {
                r_height = 3*r_width/4;
              }
            }
          } else {
            logmesg("%s requires an argument\n", argv[apc]);
          }
          break;

        case '?':
          fprintf(stderr, USAGE);
          logmesg(USAGE);
          break;

        case 'M': /* Skip the title */
          no_title=1;
          break;

        default: /* no - option, use this argument as the command filename */
          if(cmd_filename==NULL) {
            cmd_filename=argv[apc];
          } else {
            fprintf(stderr, USAGE);
            logmesg("bad option: %s\n", argv[apc]);
            logmesg(USAGE);
          }
          break;
      } /* switch */
    } /* if - */
  } /* for apc... */


  /* Open the command script file */
  if((cmd_filename==NULL) || !strcmp(cmd_filename, "-")) {
    in_fp = stdin;
  } else {
    if((in_fp=fopen(cmd_filename, "r"))==NULL) {
      logmesg("\n%s: could not open command file %s\n",
              argv[0], cmd_filename);
      reset_display();
      bye_bye(1);
    }
  }

  /* Read the command file into memory unless commands come from stdin */
  if(in_fp!=stdin) {
    load_text_file(in_fp, &cmd_script);
    fclose(in_fp);
    cmd_script_active=TRUE;
  }

  /* Set up the graphics screen */
  initialize_display(r_width, r_height);

  /* Initialize the bitmapped fonts */
  init_font();

  /* Install the "CGA" colormap */
  install_cmap(/*num*/16, /*palette*/cga_color, /*maxval*/255);

  /* Show the title screen */
  if(!no_title) {
    load_pgm_from_c(pgm_cols_title, pgm_rows_title, &sd, pgm_image_title);
    gm_xscale = (float)w_width/pgm_cols_title;
    gm_yscale = (float)w_height/pgm_rows_title;
    show_image(&sd);
    st_pause(-1.0);
    gm_xscale = 1;
    gm_yscale = 1;
  }

  /*-=-* MAIN LOOP *-=-*/

  /* Play frames and execute commands from the script */
  while(gm_play_mode.stop==MODE_GO  &&  in_fp!=NULL) {

    /* Play a frame or execute a command from the script */
    if(gm_play_mode.play==MODE_PLAY) {
      /* Continuous play */
      /* Don't bother to advance frame if primary image file is not open */
      if(FIOD[0]!=NULL) {
        if(gm_step!=0.0) pause_init();
        /* execute a faux :next command */
        if(parse_simple(":next", ":next\n", 5, &sd)==EOF) {
          /* Reached end of file */
          /* If this was not a single step, close image file */
          if(gm_play_mode.step != MODE_STEP) {
            parse_simple(":close", ":close\n", 6, &sd);
          }
        }
        pause_finish(gm_step);
      }
      if(FIOD[0] == NULL) {
        /* reset play mode to pause, no step, execute commands */
        gm_play_mode.play = MODE_PAUSE;
        gm_play_mode.step = MODE_STAY;
        gm_play_mode.commands = MODE_COMMANDS;
      }

    } else if(gm_play_mode.play==MODE_PAUSE) {
      if(gm_play_mode.step == MODE_STEP) {
        /* Play one frame */
        /* Don't bother to advance frame if primary image file is not open */
        if(FIOD[0] != NULL) {
          /* execute a faux :next command */
          if(parse_simple(":next", ":next\n", 5, &sd)==EOF) {
            /* Reached end of file, but ignore it and stay in image file*/
          }
        }
        gm_play_mode.step = MODE_STAY;
        gm_play_mode.commands = MODE_NO_COMMANDS;

      } else if(gm_play_mode.commands==MODE_COMMANDS) {
        /* Read and execute a command */

        /* Read a command line */
        if(cmd_script_active) {
          /* Read command from script in memory (if there is a cmd to read) */
          if(   (cmd_index+cmd_direction>=0)
              && cmd_script[cmd_index+cmd_direction]!=NULL )
          {
            /* Advance cmd script index */
            cmd_index += cmd_direction;
            /* copy command line into a working buffer */
            strcpy(cmd_line, cmd_script[cmd_index]);
          } else {
            /* Update the cmd_index to just beyond the valid range */
            /* ...so when we change directions, we get the next cmd correct */
            if(cmd_index+cmd_direction == -1) {
              cmd_index = -1;
            }  else if(cmd_script[cmd_index+cmd_direction]==NULL) {
              cmd_index += cmd_direction;
            }
            /* Null out the command */
            cmd[0] = cmd_line[0] = '\0';
            /* Do not accept more commands (go into playback control mode) */
            gm_play_mode.commands=MODE_NO_COMMANDS;
          }

        } else {
          /* Read command from file */
          if(fgets(cmd_line, 255, in_fp)==NULL) {
            /* Reached end of file */
            in_fp=NULL;
            /* Null out the command */
            cmd[0] = cmd_line[0] = '\0';
            /* Do not accept more commands (go into playback control mode) */
            gm_play_mode.commands=MODE_NO_COMMANDS;
          }
        }

        /* Execute a command */
        if(gm_play_mode.stop!=MODE_STOP && in_fp!=NULL) {
          /* Extract the command string from the command line */
          scan_items=sscanf(cmd_line, "%s%n", cmd, &cmd_len);
          if(scan_items<1) cmd[0]='\0';

          /* Pre-parse backward-running commands */
          if(cmd_direction < 0) {
            pre_parse(cmd, cmd_line, cmd_script, &cmd_index);
          }

          if(parse_simple(cmd, cmd_line, cmd_len, &sd)!=1) {
            /* Parse a "simple" command */

          } else if(STREQ(":xform", cmd)) {
            /* Record an xform */
            /* Free old one if it's not empty */
            if(gm_xform_script[0]!=NULL) {
              free_block(gm_xform_script);
            }
            /* Load new xform from memory or file */
            /* (skip the ":xform" by adding 1 to cmd_index) */
            cmd_index = copy_block(cmd_index+1, ":endxform",
                cmd_script, gm_xform_script);

            /* Point command interpretter to next command after :endxform */
            if(cmd_index<0) {
              cmd_index = -cmd_index;
              logmesg("reached end of commands before :endxform\n");
            } else {
              /* Refer to next command in list */
              /*cmd_index += cmd_direction;/**/
            }

          } else if(STREQ(":exit", cmd)) {
            /* Exit current command stream */
            in_fp=NULL; /* Exit main loop flag value */
            gm_play_mode.stop=MODE_STOP;

          } else if(strlen(cmd)) {
            /* (image filename);  Display entire image file */
            /* (This must go last in the parse since it accepts any string) */

            /* create the faux :open command */
            sprintf(cmd_line, ":open %s\n", cmd);
            if(parse_simple(":open", cmd_line, 5, &sd)==EOF) {
              logmesg("play file '%s' at %i did not work\n", cmd, cmd_index);
              gm_play_mode.play=MODE_PAUSE;
            } else {
              /* set continuous play mode */
              gm_play_mode.play=MODE_PLAY;
            }

          }
          /* else command was blank, ignore */

        } /* if(command in buffer) command execution block */
      } /* pause-mode if block */
    } /* play_mode if block */

    /* Handle keyboard an mouse events */
    gm_play_mode=handle_events(gm_play_mode);
    cmd_direction = (gm_play_mode.direction == MODE_FORWARD) ? 1 : -1;

  } /* while(go-mode) input and execution loop */

  /* Clean up the display */
  reset_display();

  /* Free up allocated image memory */
  free_movie(&sd);

  /* Bye bye */
  bye_bye(0);
  return(0);
}
